"""
Storage for points data

Josselin Duchateau, IHU liryc
Last update 06/12/2021
"""
from collections.abc import MutableMapping
from enum import IntEnum
from xml.etree import ElementTree as etree
from .utils import LazyClass, load_pv
from typing import Union, Iterator, Tuple, Optional, Any
import re
import numpy as np


class PointType(IntEnum):
    """ Enumeration of different point types """
    NORMAL = 0
    LOCATION_ONLY = 1
    SCAR = 2
    FLOATING = 3
    TRANSIENT_EVENT = 4


class SigReference:
    """ Signal reference """
    def __init__(self, id, ts, ref_ts_delay, lat_ts_delay, woi, uni, bi, uni_chan, bi_chan, ref_chan):
        self.id = id
        self.ts = ts  # PIU_START_TIME
        self.ref_ts_delay, self.lat_ts_delay = ref_ts_delay, lat_ts_delay # REF_ANNOTATION_TIME_DATA, MAP_ANNOTATION_LAT
        self.woi = woi  # MAP_ANNOTATION_WOI_FROM, MAP_ANNOTATION_WOI_TO
        self.uni, self.bi = uni, bi  # VOLTAGE_UNIPOLAR, VOLTAGE_BIPOLAR
        self.uni_chan, self.bi_chan, self.ref_chan = uni_chan, bi_chan, ref_chan
        # MAP_ELECTRODE1_INDEX # MAP_ANNOTATION_CHANNEL_INDEX, # REF_ANNOTATION_CHANNEL_INDEX

    def __repr__(self):
        ref_info = f' ref: {self.ref_chan}' if self.ref_chan else ''
        return f'<Signal Reference: @ts {self.ts}, channels: {self.uni_chan} | {self.bi_chan}{ref_info}>'


class LazySigRef(LazyClass, SigReference):
    """ Lazy loader for a signal reference """
    def __init__(self, data_source, filename):
        LazyClass.__init__(self)
        self._data_source = data_source
        self._filename = filename

    @property
    def _lazy(self):
        return ['id', 'ts', 'ref_ts_delay', 'lat_ts_delay', 'woi',
                'uni', 'bi', 'uni_chan', 'bi_chan', 'ref_chan']

    @property
    def filename(self):
        """ Filename to load when properties are accessed"""
        return self._filename

    def load(self):
        """ Load the target filename """
        with self._data_source.open(self._filename) as f:
            root = etree.parse(f).getroot()
            self.id = root.get('ID')
            # self.date = root.get('Date')
            # self.time = root.get('Time')

            # Annotations
            annotations = root.find('Annotations')
            self.ts = int(annotations.get('StartTime'))
            ref_ts_delay = int(annotations.get('Reference_Annotation'))
            self.ref_ts_delay = ref_ts_delay
            lat_ts_delay = int(annotations.get('Map_Annotation'))
            if lat_ts_delay - ref_ts_delay == - 10000:
                self.lat_ts_delay = np.nan
            else:
                self.lat_ts_delay = lat_ts_delay

            # WOI
            woi = root.find('WOI')
            self.woi = (int(woi.get("From")), int(woi.get("To")))

            # Voltages
            voltages = root.find('Voltages')
            self.uni = float(voltages.get("Unipolar"))
            self.bi = float(voltages.get("Bipolar"))

            # ECG datafile
            ecg = root.find('ECG')
            ecg_filename = ecg.get('FileName')

        # Read the headers of the ecg_filename
        with self._data_source.open(ecg_filename) as f:
            f.readline()  # Skip headers
            f.readline()
            mapping_chan_line = f.readline().decode('utf-8')
        mapping_chan = re.match(r'Unipolar Mapping Channel=(?P<uni>[\w-]+) '
                                r'Bipolar Mapping Channel=(?P<bi>[\w-]+)'
                                r'( Reference Channel=(?P<ref>[\w-]+))?', mapping_chan_line)

        uni_chan, bi_chan, ref_chan = mapping_chan['uni'], mapping_chan['bi'], mapping_chan['ref']
        self.uni_chan = uni_chan.replace('_', ' ')
        self.bi_chan = bi_chan.replace('_', ' ')
        self.ref_chan = ref_chan.replace('_', ' ')

        self._loaded = True

    def __repr__(self):
        if self._loaded:
            ref_info = f' ref: {self.ref_chan}' if self.ref_chan else ''
            return f'<Signal Reference: @ts {self.ts}, channels: {self.uni_chan} | {self.bi_chan}{ref_info}>'
        return '<Signal Reference (lazy load)>'


class Point(MutableMapping):
    """ A class describing a point with annotations +/- signals """

    def __init__(self, id: str, x: float, y: float, z: float, a: float = 0., b: float = 0., g: float = 0.,
                 cath_id: Optional[int] = None, sig_ref: Union[SigReference, LazySigRef] = None, **annotations):
        """
        Create an instance of a Point object.

        :param id: point id (string)
        :param x: x position (float)
        :param y: y position (float)
        :param z: z position (float)
        :param a: angle alpha (float)
        :param b: angle beta (float)
        :param g: angle gamma (float)
        :param cath_id: cathetr ID (integer)
        :param sig_ref: reference to signal (LazySigRef or SigReference object)
        :param annotations: annotations of this point (lat, voltages, etc.)
        """
        self.id = id
        self.cath_id = cath_id
        self.x, self.y, self.z = float(x), float(y), float(z)
        self.a, self.b, self.g = float(a), float(b), float(g)
        self._ref = sig_ref

        # Add the annotations to the point
        self._annotations = {}
        for annotation, value in annotations.items():
            self[annotation] = value

    @classmethod
    def from_line(cls, line, point_tags=None):
        """ Create a point from a line in a .car file
        :param line: line of text to parse in the car file
        :param point_tags: point tags from which to get tag information
        :return: the created point object
        """
        components = line.split("\t")
        pt_id = components[2]
        cath_id = int(components[19])
        xyzabg = [float(i) for i in components[4:10]]
        annotations = {
            'unipolar': float(components[10]),
            'bipolar': float(components[10]),
            'lat': int(components[12]),
            'impedance': float(components[13]),
            'point type': int(components[14]),
            'tag': point_tags(int(components[15])) if point_tags is not None else int(components[15])
        }
        if annotations['lat'] == -10000:
            annotations['lat'] = np.nan
        return cls(pt_id, *xyzabg, cath_id=cath_id, **annotations)

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise KeyError('Point-related data attributes are stored as strings')
        try:
            return self._annotations.__getitem__(item)
        except KeyError:
            return np.nan

    def __setitem__(self, key, value) -> None:
        if not isinstance(key, str):
            raise KeyError('Point-related data attributes are stored as strings')
        return self._annotations.__setitem__(key, value)

    def __delitem__(self, key) -> None:
        return self._annotations.__delitem__(key)

    def __len__(self) -> int:
        return self._annotations.__len__()

    def __iter__(self) -> Iterator:
        return self._annotations.__iter__()

    def __repr__(self) -> str:
        return f"<Point id='{self.id}' @ {self.x}, {self.y}, {self.z}>"

    def __copy__(self):
        """
        Create a copy of the point

        :return: New point with copied properties and annotations
        """
        point_copy = Point(self.id, self.x, self.y, self.z, self.a, self.b, self.g, self._ref)

        # Copy annotations
        for annotation, value in self.items():
            point_copy[annotation] = value

        return point_copy

    @property
    def pos(self) -> Tuple[float, float, float]:
        """ Position getter """
        return self.x, self.y, self.z

    @pos.setter
    def pos(self, value) -> None:
        """ Position setter """
        self.x, self.y, self.z = float(value[0]), float(value[1]), float(value[2])

    @property
    def angle(self) -> Tuple[float, float, float]:
        """ Angle getter """
        return self.a, self.b, self.g

    @angle.setter
    def angle(self, value) -> None:
        """ Angle setter """
        self.a, self.b, self.g = float(value[0]), float(value[1]), float(value[2])

    @property
    def ref(self):
        """ Reference to the signal - getter """
        if isinstance(self._ref, LazySigRef) and not self._ref.loaded:
            self._ref.load()
        return self._ref

    @ref.setter
    def ref(self, value):
        """ Reference to the signal - setter """
        if not isinstance(value, SigReference):
            raise ValueError('Signal reference must be a valid reference object.')
        self._ref = value


class PointSet(MutableMapping):
    """ A point collection """
    def __init__(self, name, points:dict =None):
        self._name = name
        self._points = points if points else {}

    @classmethod
    def from_car_file(cls, data_source, filename, point_tags=None):
        """ Read point collection from .car file"""
        with data_source.open(filename) as car_file:
            line = car_file.readline().decode('utf-8')
            #header = re.match(r'VERSION_(?P<version>[\d_])+ (?P<map_name>[^\r\n]+)', line)
            #name = header['map_name']
            # version = header['version'].replace('_', '.')
            name = re.match(r'(?P<study>.*)_car\.txt', filename)['study']
            points = {}
            for line in car_file:
                pt = Point.from_line(line.decode('utf-8'), point_tags)
                pt.ref = LazySigRef(data_source, f'{name}_P{pt.id}_Point_Export.xml')
                points[pt.id] = pt
        return cls(name, points)

    @classmethod
    def from_db(cls, db_cursor, map_name):
        # TO DO: complete code
        """ Read point collection from database """
        query = f"SELECT MAP_IDX FROM MAPS_TABLE WHERE NAME={map_name}"
        db_cursor.execute(query)
        try:
            idx = next(db_cursor)[0]
        except StopIteration:
            raise ValueError('Could not find specified map in maps tables.')

        query = (f"SELECT ("
                 f"POINT_IDX, X_COORDINATE, Y_COORDINATE, Z_COORDINATE, "
                 f"ALPHA_COORDINATE, BETA_COORDINATE, GAMMA_COORDINATE, "
                 f"VOLTAGE_UNIPOLAR, VOLTAGE_BIPOLAR, "
                 f"MAP_ANNOTATION_LAT, IMPEDANCE,"
                 f"POINT_TYPE, TAG_TYPE, CATHETER_ID"
                 f") FROM POINTS_TABLE WHERE MAP_IDX={idx}")

        db_cursor.execute(query)
        points = {}
        for row in db_cursor:
            pt = Point(*row)
            points[pt.id] = pt
            # HERE WE NEED TO CREATE A REFERENCE TO THE BINARY DATA ?

        cls(map_name, points)

    @property
    def name(self):
        """ Name of the point collection """
        return self._name

    def duplicate(self, new_name: str = None):
        """
        Create a copy of the point set. Each individual point is copied as a new instance.

        :param new_name: New name for the copied set
        :return: A copy of the point set with duplicated points
        """
        new_name = new_name or self._name + ' - duplicate'
        duplicate = PointSet(new_name)

        for pt in self._points.values():
            pt_copy = pt.__copy__()
            duplicate[pt_copy.id] = pt_copy

        return duplicate

    def attribute_array(self, attribute) -> np.ndarray:
        """
        Return a numpy array of the requested attribute

        :param attribute: Attribute
        :return: Numpy array
        """
        return np.asarray([pt[attribute] for pt in self.values()])

    def project(self, mesh, threshold: Optional[float] = None) -> Any:
        """
        Project the point set to a given mesh.

        :param mesh: Carto mesh or pyvista mesh on which we should project the points. The points are projected to the
        nearest cell on this mesh.
        :param threshold: Optional threshold distance above which points should be discarded
        """
        # Duplicate points and get the position array
        duplicate = self.duplicate(self._name + '-Projected')
        point_positions = duplicate.position_array

        # Get a pyvista version of the mesh
        pv_mesh = mesh.pv_mesh(cutouts=False)

        # Project the points to the nearest cell, store the distance
        index, projections = pv_mesh.find_closest_cell(point_positions, True)
        distances = np.sqrt(np.sum((point_positions - projections) ** 2, 1))
        for pt, distance in zip(duplicate.values(), distances):
            pt['projection distance'] = distance

        # Filter points if necessary
        if threshold is not None:
            for pt_id, distance in zip(list(duplicate), distances):
                if distance > threshold:
                    del duplicate[pt_id]

        return duplicate

    @property
    def position_array(self):
        """
        Get a numpy array with all the point positions.

        :return:
        """
        return np.asarray([pt.pos for pt in self.values()])

    def __getitem__(self, item):
        if isinstance(item, int):
            item = list(self.keys())[item]
        return self._points.__getitem__(item)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            key = list(self.keys())[key]
        if isinstance(key, str):
            if not key.isnumeric() or int(key) < 1:
                raise TypeError('Keys must be a string representation of a positive integer.')
        return self._points.__setitem__(key, value)

    def __delitem__(self, key):
        if isinstance(key, int):
            key = list(self.keys())[key]
        return self._points.__delitem__(key)

    def __len__(self):
        return self._points.__len__()

    def __iter__(self):
        return self._points.__iter__()

    def __repr__(self):
        return f"<Point Set '{self.name}': {len(self._points)} points>"


import sys
import os.path
import csv
import pyodbc


# 1) Connexion to MSSQL Server: specify the absolute path to the driver library.
# Example of driver file path:
# "Driver={/opt/microsoft/msodbcsql17/lib64/libmsodbcsql-17.7.so.2.1};"
driver = "Driver={/usr/local/lib/libmsodbcsql.17.dylib};"

# 2) Input the correct uid and password
# Example:
# "Uid=Billy;"
# "Pwd=someRobustPassword;"
uid = "Uid=;"
password = "Pwd=;"

# 3) Feel free to add or remove tables
tables_to_convert = ["MAPS_TABLE", "SETUP_TABLE", "STUDIES_TABLE", "TAG_SETUP_TABLE",
                     "PROCEDURES_TABLE", "POINTS_TABLE", "COLORS_TABLE"]


# get database path and print help if necessary
def get_database_path(argv):
    if len(argv) != 2:
        print(f'Usage: {argv[0]} /absolute/path/to/database.mdf')
        quit()
    else:
        database = argv[1]
        if database.endswith(".mdf") and database.startswith('/'):
            return database
        else:
            print(f'Invalid database path or type: must pass an absolute path to a mssql server database')
            quit()


def convert_table(cursor, table_name, output_path):
    output_file_name = output_path + "/" + table_name.lower() + ".csv"
    query = """SELECT * FROM """ + table_name
    cursor.execute(query)
    column_names = [column[0] for column in cursor.description]

    os.makedirs(os.path.dirname(output_file_name), exist_ok=True)
    # Go through the results row-by-row and write the output to a CSV file.
    # Use QUOTE_NONE, could be changed to QUOTE_NONNUMERIC or
    # QUOTE_ALL (to avoid escape character problems)
    # Use '|' as delimiter, this seems to work fine for now. Will be problematic
    # if some fields contain '|' as a character.
    with open(output_file_name, "w") as outfile:
        writer = csv.writer(outfile, quoting=csv.QUOTE_NONE, delimiter='|')
        # write the column names first
        writer.writerow(column_names)
        # write the data
        for row in cursor:
            writer.writerow(row)


def convert_database(input, output_path):
    server = "Server=tcp:localhost;"
    database_name = "Database=TemporaryCartoDatabase;"
    attach_db_filename = "AttachDbFileName=" + input + ";"

    connection_string = driver + server + database_name + attach_db_filename + uid + password
    connexion = pyodbc.connect(connection_string)
    connexion.autocommit = False

    cursor = connexion.cursor()

    for table_name in tables_to_convert:
        convert_table(cursor, table_name, output_path)

    # Close the cursor and the database connection
    cursor.close()
    connexion.close()


if __name__ == '__main__':
    database_path = get_database_path(sys.argv)
    output_path = os.path.dirname(database_path) + "/converted_database"

    convert_database(database_path, output_path)


# All available tables found in the database (tested only on 1 database, there may be more):
#
# A_TAGS_TABLE
# ARCHIVE_LOG_TABLE
# AUTO_MAP_PRESETS_TABLE
# BCS_TABLE
# CARTOFINDER_SITES_TABLE
# CATHETERS_ID_TABLE
# CFAE_SETUP_TABLE
# COLOR_SCALE_SETUP_TABLE
# COLORS_TABLE
# CONFIG_ACQUIRE_TABLE
# CONFIG_CARTOFINDER_PREFERENCE_TABLE
# CONFIG_CHANNEL_TABLE
# CONFIG_CONNECTOR_TABLE
# CONFIG_CONTACT_FORCE_GUI_TABLE
# CONFIG_CONTACT_FORCE_PREFERENCE_TABLE
# CONFIG_EP_RECORDING_TABLE
# CONFIG_FILTER_TABLE
# CONFIG_HELIOS_PRESET_TABLE
# CONFIG_HUD_TOOLBAR_TABLE
# CONFIG_LAYOUT_TABLE
# CONFIG_MAIN_TABLE
# CONFIG_MAP_ANNOTATION_TABLE
# CONFIG_MAP3D_DISPLAY_TABLE
# CONFIG_MAP3D_VIEW_TABLE
# CONFIG_PANE_TABLE
# CONFIG_PASO_SETTINGS_TABLE
# CONFIG_PHASE_TABLE
# CONFIG_REF_ANNOTATION_TABLE
# CONFIG_RIPPLE_MAPPING_PREFERENCE_TABLE
# CONFIG_STIMULATION_TABLE
# CONFIG_TEMPERATURES_PREFERENCE_TABLE
# CONFIG_TGA_PRESET_TABLE
# CONFIG_VISI_TAG_PREFERENCE_TABLE
# CONFIG_VISI_TAG_PRESET_TABLE
# CONFIG_WINDOW_SETTINGS_TABLE
# CONNECTOR_FILTER_TABLE
# CS_COORD_SYSTEM_TABLE
# DATABASE_VERSION
# DICOM_DIR_TABLE
# DICOM_II_TABLE
# ECG_PATTERNS_TABLE
# ECG_PRESET_PARAMETERS_TABLE
# ECG_PRESETS_TABLE
# ECG_STUDY_PRESET_PARAMETERS_TABLE
# ECG_STUDY_WINDOW_PRESET_PARAMETERS_TABLE
# GENERIC_SETUP_TABLE
# MAP_ANNOTATION_MAP_SETUP_TABLE
# MAPS_TABLE
# PASO_IS_IS_SIGNALS_CORR_TABLE
# PASO_IS_TABLE
# PASO_PM_IS_SIGNALS_CORR_TABLE
# PASO_PM_TABLE
# PATIENTS_TABLE
# PHYSICIANS_TABLE
# POINT_GROUPS_TABLE
# POINTS_TABLE
# PRESET_FILTER_TABLE
# PROCEDURES_TABLE
# REF_ANNOTATION_MAP_SETUP_TABLE
# RF_TABLE
# SCAR_TABLE
# SECONDARY_WINDOWS_POSITIONS
# SEGMENTED_VOLUME_TABLE
# SETUP_TABLE
# STUDIES_TABLE
# STUDY_BACKUP_COMMIT_FILES_TABLE
# STUDY_BACKUP_HISTORY_TABLE
# STUDY_BACKUP_INFO_TABLE
# STUDY_RECORDINGS_TABLE
# TAG_SETUP_TABLE
# TRIAL_LICENSE_TABLE
# ULS_ANNOTATIONS_TABLE
# ULS_CLIPS_TABLE
# ULS_CONTOURS_TABLE
# ULS_FRAMES_POSITIONS_TABLE
# ULS_FRAMES_TABLE
# ULS_GATED_FRAMES_TABLE
# ULS_POINTS_TABLE
# ULS_TAGS_PER_CONTOUR_TABLE
# USED_LICENSE_TABLE
# VIEW_SETUP_TABLE
# VOLUME_TABLE



import zipfile
import os
from contextlib import contextmanager
from functools import lru_cache
from abc import ABC, abstractmethod
from typing import List
import re
import sys


class DataSource:
    """ An abstract class describing a datasource: either standard folder or a Zip file"""
    def __init__(self, path: str):
        self.path = path
        if not os.path.isdir(path):
            self._archive: zipfile.ZipFile = zipfile.ZipFile(self.path, mode='r')
        else:
            self._archive = None

    @contextmanager
    def open(self, file):
        """ Wrapper around the zipfile.open and open context managers """
        if self._archive:
            f = self._archive.open(file, 'r')
        else:
            f = open(os.path.join(self.path, file), mode='rb')
        try:
            yield f
        finally:
            f.close()

    @lru_cache(None)
    def listdir(self):
        """ List directory contents """
        if self._archive:
            return self._archive.namelist()
        else:
            return os.listdir(self.path)

    def __repr__(self):
        if self._archive:
            return f'<DataSource @ {self.path} (Zip archive)>'
        else:
            return f'<DataSource @ {self.path}>'


class LazyClass(ABC):
    """ A Lazy class that loads data on demand """
    def __init__(self):
        self._loaded = False
        for attribute in self._lazy:
            self.__setattr__(attribute, None)

    def __getattribute__(self, item):
        if item in ['_lazy', '_loaded', 'load', '__setattr__', 'loaded']:
            return super().__getattribute__(item)

        if item in self._lazy and not self._loaded:
            self.load()
        return super().__getattribute__(item)

    @abstractmethod
    def load(self):
        pass

    @property
    @abstractmethod
    def _lazy(self) -> List[str]:
        pass

    @property
    def loaded(self):
        return self._loaded


def make_valid_tag_name(var_name: str):
    """ Make a valid tag name """
    split = re.split(r'[^0-9a-zA-Z_]', var_name)
    var_name = '_'.join([cp.upper() for cp in split if cp])
    return re.sub('^[^a-zA-Z_]+', '', var_name)


def in_notebook():
    """ Check to see if we are in jupyter notebook """
    return 'ipykernel' in sys.modules


def load_pv():
    """ Load pyvista """
    try:
        import pyvista as pv
    except ImportError:
        raise ImportError('Pyvista must be installed for this method to work. '
                          'Use pip install carto_reader[viz] to install visualization dependencies.')
    return pv



import os
import pyodbc
import docker

MSSQL_DRIVER_PATH = os.getenv('MSSQL_DRIVER_PATH', '/usr/local/lib/libmsodbcsql.17.dylib')
MSSQL_SA_PASSWORD = 'Adm1nP@55w0rd'


if not os.path.exists(MSSQL_DRIVER_PATH):
    raise EnvironmentError('MSSQL_DRIVER_PATH environment variable is not correct.')


def create_docker_mssql(db_path):
    mount_path = os.path.abspath(db_path)
    if not all(file in os.listdir(mount_path) for file in ['EmptyPack_Data.mdf', 'EmptyPack_Log.ldf']):
        raise FileNotFoundError('Count not find the carto database files in the specified path.')

    mount = docker.types.Mount(target='/var/opt/mssql/data/', source=mount_path, type='bind')
    client = docker.from_env()
    container = client.containers.run('mcr.microsoft.com/mssql/server:2017-CU8-ubuntu',
                                      environment={'ACCEPT_EULA': 'Y', 'SA_PASSWORD': MSSQL_SA_PASSWORD},
                                      ports={1433: 1433},
                                      mounts=[mount],
                                      detach=True)
    return container



class Carto:
    def __init__(self, path):
        """
        Initialize the object
        :param path: the path to the MDF database file to open
        """
        self._path = os.path.abspath(path)
        connection_args = {
            'Driver': MSSQL_DRIVER_PATH,
            'Server': 'localhost,1433',
            'Database': 'TemporaryCartoDatabase',
            'AttachDbFileName': '{/var/opt/mssql/data/EmptyPack_data.mdf}',
            'Uid': 'sa',
            'Pwd': 'Adm1nP@55w0rd'
        }
        cx = pyodbc.connect(''.join([f'{k}={v};' for k, v in connection_args.items()]), autocommit=False)
        cursor = cx.cursor()

        print([row.table_name for row in cursor.tables()])

        cx.close()




"""
Signal storage for carto data

Josselin Duchateau, IHU Liryc
Last update 06/12/2021
"""
import numpy as np

CARTO_SIGNAL_FILE_LENGTH = 2500


class Channel:
    """ Class containing ECG channel data. The channel behaves (more or less) like a numpy array. """
    def __init__(self, name):
        self._signals = {}
        self._name = name

    def __getitem__(self, item):
        if isinstance(item, slice):
            start = item.start or self.first_sample
            stop = item.stop or self.last_sample
            step = item.step or 1

            if start > stop:
                raise ValueError('Start must be before stop')

            pos = start
            parts = []
            for itv in self.signal_parts:
                if pos < itv[0]:
                    end = min(itv[0], stop)
                    length = 1 + ((end - pos - 1) // step)
                    parts.append(np.full((length, ), np.nan))
                    pos += length * step

                if itv[0] <= pos < itv[1]:
                    end = min(itv[1], stop)
                    length = 1 + ((end - pos - 1) // step)
                    parts.append(self._signals[itv][(pos  - itv[0]):(end  - itv[0]):step])
                    pos += length * step

                if pos >= stop:
                    break

            # Final NaNs
            if pos < stop:
                length = 1 + ((stop - pos - 1) // step)
                parts.append(np.full((length, ), np.nan))
                pos += length * step

            signal = np.concatenate(parts)
            return signal

    @property
    def first_sample(self):
        """ First sample of the channel """
        return min((start for start, _ in self._signals))

    @property
    def last_sample(self):
        """ Last sample of the channel """
        return max((stop for _, stop in self._signals))

    @property
    def signal_parts(self):
        """ Parts of the channel that contain signal """
        return sorted(self._signals)

    def add(self, signal, start_ts: int):
        """ Add a signal to the channel """
        itv = start_ts, len(signal) + start_ts
        if itv not in self._signals:
            self._signals[itv] = signal

        self.simplify_parts()

    def simplify_parts(self):
        """ Simplify the parts (merging them if necessary) """
        prev_itv = (-1, -1)
        prev_signal = None
        new_parts = {}
        for itv in self.signal_parts:
            if itv[0] <= prev_itv[1]:  # Overlap
                if itv[1] <= prev_itv[1]:  # We can remove this part
                    continue
                new_itv = prev_itv[0], itv[1]
                new_signal = np.concatenate((prev_signal, self._signals[itv][prev_itv[1]-itv[0]:itv[1]]))
                prev_itv = new_itv
                prev_signal = new_signal
                continue
            else:  # No overlap
                if prev_signal is not None:
                    new_parts[prev_itv] = prev_signal
                prev_itv = itv
                prev_signal = self._signals[itv]

        if prev_signal is not None:
            new_parts[prev_itv] = prev_signal

        self._signals = new_parts

    def defined(self, start, stop):
        """ Check if the signal is defined in the given interval """
        if start >= stop:
            raise ValueError('Invalid start/stop bounds')
        for begin, end in self.signal_parts:
            if begin <= start and stop <= end:
                return True
        return False

    @property
    def name(self):
        """ Name of the channel """
        return self._name

    @property
    def total_data_length(self):
        """ Total length of the data (non-NaN)"""
        return sum([e-s for s, e in self.signal_parts])

    @property
    def acquisition_time_span(self):
        """ Total acquisition time (from first to last defined sample)"""
        return self.signal_parts[-1][1] - self.signal_parts[0][0]

    @property
    def num_parts(self):
        """ Nuber of signal parts (continuous data strips)"""
        return len(self.signal_parts)

    def __repr__(self):
        return f'<Channel {self.name}: {self.total_data_length//1000} s of signal in ' \
               f'{self.num_parts} parts over {self.acquisition_time_span//1000} s>'



"""
Mesh information

Josselin Duchateau, IHU Liryc
Last updated 06/12/2021
"""
from dataclasses import dataclass
import numpy as np
from collections.abc import MutableMapping
from typing import List, Optional
from .utils import load_pv

# Colormaps for the default mesh maps
DEFAULT_COLORMAPS = {
    'unipolar': ('gist_rainbow', [1.5, 5]),
    'bipolar': ('gist_rainbow', [0.5, 1.5]),
    'lat': ('hsv', None),
    'groupid': ('tab20', None)
}


@dataclass
class MeshAttributes:
    """ Attributes of the mesh """
    MeshID: int
    MeshName: str
    NumVertex: int
    NumTriangle: int
    TopologyStatus: int
    MeshColor: List[float]
    Matrix: List[float]
    NumVertexColors: int
    ColorsIDs: Optional[List[int]] = None
    ColorsNames: Optional[List[str]] = None
    NumVertexInitial: int = None
    NumTriangleInitial: int = None
    NumVisibleGroups: int = None
    VisibleGroupsIDs: List[int] = None
    VisibleGroupsTypes: List[int] = None
    NumTransparentGroups: int = None
    TransparentGroupsIDs: List[int] = None
    TransparentGroupsTypes: List[int] = None


class Mesh(MutableMapping):
    """ A class containing Carto FAM meshes and colormaps"""
    def __init__(self, vertices: np.ndarray = None, triangles: np.ndarray = None,
                 vertice_normals: np.ndarray = None, triangle_normals: np.ndarray = None,
                 attributes=None):
        self.v = vertices
        self.t = triangles
        self.vn = vertice_normals
        self.tn = triangle_normals

        self.attributes = attributes
        self._maps = {}

    @classmethod
    def from_file(cls, data_source, filename):
        """ Read a mesh from a file """
        if filename not in data_source.listdir():
            raise ValueError('Requested mesh not found in datasource.')

        def listify(value_string):
            """ Turn into a list"""
            item_list = value_string.strip().split()
            if not item_list:
                return None

            try:
                item_list = [int(item) for item in item_list]
            except ValueError:
                try:
                    item_list = [float(item) for item in item_list]
                except ValueError:
                    pass
            if len(item_list) == 1:
                return item_list[0]

            return item_list

        def read_attributes(_f):
            """ Read the attributes section """
            attributes = {}
            for _line in _f:
                _line = _line.decode('latin-1').rstrip()
                if not _line:
                    break
                if _line[0] == ';':
                    continue
                name, value = _line.split('=')
                attributes[name.strip()] = listify(value)
            return MeshAttributes(**attributes)

        def line_filter(_f):
            """ Filter line """
            got_data = False
            for _line in _f:
                if not _line.rstrip():
                    if got_data:
                        return
                    else:
                        continue
                if _line[0:1] == b';':
                    continue
                _line = _line.replace(b'=', b' ')
                got_data = True
                yield _line

        obj = cls()
        maps = None
        v_attributes = None
        with data_source.open(filename) as f:
            for line in f:
                line = line.rstrip()
                if line == b'[GeneralAttributes]':
                    obj.attributes = read_attributes(f)
                if line == b'[VerticesSection]':
                    vertices = np.loadtxt(line_filter(f))
                    obj.v = vertices[:, 1:4]
                    obj.vn = vertices[:, 4:7]
                    # obj._maps['vGroupID'] = vertices[:, 7]  # Vertex group info doesn't contain any data usually...
                if line == b'[TrianglesSection]':
                    triangles = np.loadtxt(line_filter(f))
                    obj.t = triangles[:, 1:4].astype(int)
                    obj.tn = triangles[:, 4:7]
                    group_id = -triangles[:, 7].astype(int)
                    group_id[group_id == 1000000] = -1
                    obj._maps['GroupID'] = group_id
                if line == b'[VerticesColorsSection]':
                    maps = np.loadtxt(line_filter(f))
                    maps[maps == -10000.] = np.nan
                if line == b'[VerticesAttributesSection]':
                    v_attributes = np.loadtxt(line_filter(f), dtype=int)
                    if len(v_attributes) == 0:
                        v_attributes = None

        if obj.attributes.ColorsNames is not None and maps is not None:
            for map, map_name in zip(maps.transpose()[1:], obj.attributes.ColorsNames):
                obj._maps[map_name] = map
        if v_attributes is not None:
            obj._maps['Scar'] = v_attributes.transpose()[1]
            obj._maps['EML'] = v_attributes.transpose()[2]
        return obj

    def interpolate_points(self, point_set, attribute: str, map_name: Optional[str] = None,
                           interpolate_nan: bool = False, **interpolation_options):
        """
        Create a new map on the mesh by interpolating points
        :param point_set: the point set from which the data must be interpolated
        :param attribute: the attribute to interpolate
        :param map_name: optional name of the map to create (defaults to the attribute name)
        :param interpolate_nan: interpolate NaN values or remove them before hand
        :param interpolation_options: interpolation options to pass to the interpolate function.
        These options include: radius, sharpness, strategy, null_value, n_points, etc.
        See the pyvista.DataSetFilters.interpolate documentation for details.
        """
        pv = load_pv()
        map_name = map_name or attribute

        # Get position and attribute arrays
        pos = point_set.position_array
        vals = point_set.attribute_array(attribute)

        # Filter NaN Values
        if not interpolate_nan:
            pos = pos[np.logical_not(np.isnan(vals))]
            vals = vals[np.logical_not(np.isnan(vals))]

        # Create pyvista version of the points/attributes data
        pts = pv.PolyData(pos)
        pts['val'] = vals

        # Get a pyvista version of our mesh, and interpolate the points to this mesh
        if not interpolation_options: # Default options if none are present
            interpolation_options = {'radius': 14, 'sharpness': 5, 'null_value': np.nan}
        pv_mesh = self.pv_mesh(cutouts=False)
        pv_mesh = pv_mesh.interpolate(pts, **interpolation_options)
        self[map_name] = pv_mesh['val']

    @property
    def num_v(self):
        """ Number of vertices """
        return len(self.v)

    @property
    def num_t(self):
        """ Number of triangles (simplexes) """
        return len(self.t)

    def __len__(self):
        return self._maps.__len__()

    def __getitem__(self, item):
        if isinstance(item, int):
            item = list(self.keys())[item]
        return self._maps.__getitem__(item)

    def __delitem__(self, key):
        if isinstance(key, int):
            key = list(self.keys())[key]
        return self._maps.__delitem__(key)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            key = list(self.keys())[key]
        if not isinstance(value, np.ndarray):
            _value = np.ndarray(value)
        else:
            _value = value
        _value.flatten()
        if len(value) != self.num_v | len(value) != self.num_t:
            raise ValueError('Input map has an invalid number of values')
        return self._maps.__setitem__(key, value)

    def __iter__(self):
        return self._maps.__iter__()

    def __repr__(self):
        return f'<Mesh ({self.num_v} x {self.num_t}), {len(self)} Maps>'

    def pv_mesh(self, cutouts=True, all_maps=False):
        """
        Get a pyvista mesh version of the current Mesh object.
        Requires a working installation of pyvista.
        :param cutouts: remove the cutouts
        :param all_maps: add all maps as scalar fields to the object
        :return:
        """
        pv = load_pv()
        if 'GroupID' in self:
            if cutouts:
                t = self.t[self['GroupID'] == 0]
            else:
                t = self.t[self['GroupID'] != -1]
        else:
            t = self.t

        poly_data = pv.PolyData(self.v, np.hstack((np.full((len(t), 1), 3, int), t)))
        if all_maps:
            for map_name, map_data in self.items():
                if len(map_data) == self.num_t and 'GroupID' in self:
                    if cutouts:
                        map_data = map_data[self['GroupID'] == 0]
                    else:
                        map_data = map_data[self['GroupID'] != -1]

                poly_data[map_name] = map_data

        return poly_data

    def plot(self, _map=None, cutouts=True):
        """ Plot the mesh
        :param _map: The map to plot (optional) - can be either a map name/index or a numpy array
        :param cutouts: remove cutouts (default is True) or display the full mesh
        :return:
        """
        pv = load_pv()
        poly_data = self.pv_mesh(cutouts)
        plotter = pv.Plotter()
        i = -1
        cmap, clim = None, None
        if _map is None:
            _map = np.zeros((self.num_v,))
        elif isinstance(_map, (str, int)):
            if isinstance(map, int):
                _map = list(self.keys())[_map]
            cmap, clim = DEFAULT_COLORMAPS.get(_map.lower(), (None, None))
            _map = self[_map]

        if len(_map) == self.num_t and 'GroupID' in self:
            if cutouts:
                _map = _map[self['GroupID'] == 0]
            else:
                _map = _map[self['GroupID'] != -1]

        curr_mesh = plotter.add_mesh(poly_data, scalars=_map, cmap=cmap, clim=clim)
        curr_text = plotter.add_text('')

        def change_map(increment):
            """ Callback function to change map """
            def plot_map():
                nonlocal i, curr_text, curr_mesh
                i = (i + increment + len(self)) % len(self)
                map_name = list(self.keys())[i]
                cmap, clim = DEFAULT_COLORMAPS.get(map_name.lower(), (None, None))
                plotter.remove_actor(curr_mesh)
                _map = self[i]

                if len(_map) == self.num_t and 'GroupID' in self:
                    if cutouts:
                        _map = _map[self['GroupID'] == 0]
                    else:
                        _map = _map[self['GroupID'] != -1]

                curr_mesh = plotter.add_mesh(poly_data, scalars=_map, cmap=cmap, clim=clim)
                plotter.remove_actor(curr_text)
                curr_text = plotter.add_text(map_name)
            return plot_map

        plotter.add_key_event('Left', change_map(-1))
        plotter.add_key_event('Right', change_map(1))
        plotter.show()



from collections.abc import Mapping
from .utils import DataSource, LazyClass, make_valid_tag_name
from .mesh import Mesh
from .points import PointSet, Point
from .signals import Channel, CARTO_SIGNAL_FILE_LENGTH
import numpy as np
import re
from tqdm import tqdm
from xml.etree import ElementTree as etree
from enum import IntEnum
import logging

logger = logging.getLogger('Carto')


# Default point tags
class PointTags(IntEnum):
    NO_TAG = -1
    NONE = 4
    HIS = 5
    PACING_SITE = 6
    DOUBLE_POTENTIAL = 7
    FRAGMENTED_SIGNAL = 8
    ABLATION = 9
    SCAR = 10
    LOCATION_ONLY = 11
    TRANSIENT_EVENT = 12


class Map(LazyClass):
    def __init__(self, data_source, map_name, point_tags=None, mesh_filename=None):
        super().__init__()

        # Initialize the mesh and point list by reading the .mesh and .car files
        self._map_name = map_name
        self._data_source = data_source
        self._mesh_filename = mesh_filename or map_name + '.mesh'
        self._point_tags = point_tags or PointTags

    def load(self):
        self._mesh = Mesh.from_file(self.data_source, self._mesh_filename)
        self._points = PointSet.from_car_file(self.data_source, f'{self.name}_car.txt', self._point_tags)

        self._loaded = True

    @property
    def name(self):
        return self._map_name

    @property
    def points(self):
        return self._points

    @property
    def mesh(self):
        return self._mesh

    @property
    def data_source(self):
        return self._data_source

    @property
    def _lazy(self):
        return ['_mesh', '_points']

    def __repr__(self):
        if self._loaded:
            return f'<Map {self.name}: {len(self.points)} Points, {self.mesh.num_v} x {self.mesh.num_t} Mesh>'
        else:
            return f'<Map {self.name} (Lazy Loaded)>'


class Carto(Mapping):
    def __init__(self, path):
        self._data_source = DataSource(path)

        # Get the study, point and point list xml files
        _xmls = {file for file in self.data_source.listdir() if file.endswith('.xml') and not '/' in file}
        _points_xml = {file for file in _xmls if file.endswith('_Point_Export.xml')}
        _point_list_xml = {file for file in _xmls if file.endswith('_Points_Export.xml')}
        _study_xml = _xmls - _points_xml - _point_list_xml
        print(_study_xml)

        if len(_study_xml) != 1:
            raise RuntimeError('Coud not find study xml file at specified path.')

        self._xmls = {
            'study': next(iter(_study_xml)),
            'points': _points_xml,
            'point_lists': _point_list_xml
        }

        # Parse the study xml file
        with self.data_source.open(self._xmls['study']) as f:
            study = etree.parse(f).getroot()
            self._name = study.get('name', None)
            tags = study.findall('Maps/TagsTable/Tag')
            if tags:
                self._tags = IntEnum('PointTags', [('NO_TAG', -1)] + [(make_valid_tag_name(tag.get('Full_Name')),
                                                                       int(tag.get('ID'))) for tag in tags])
            else:
                logger.warning('Could not get tag definition from study XML file. Using default.')
                self._tags = PointTags

            maps = study.findall('Maps/Map')
            self._maps = {
                _map.get('Name'): Map(self.data_source, _map.get('Name'), self.tags,
                                      _map.get('FileNames')) for _map in maps
                # NB: other infos: 'Volumes': mesh volume, 'Visible': visibility
            }

        # Get the list of maps from the list of car files
        # self._maps = {file[:-8]: None for file in self._data_source.listdir() if file.endswith('_car.txt')}
        self._signals = {}
        self._integrated_files = set()

    def __getitem__(self, item):
        if isinstance(item, int):
            item = list(self.keys())[item]
        if self._maps[item] is None:
            self._maps[item] = Map(self._data_source, item)
        return self._maps.__getitem__(item)

    def __iter__(self):
        return self._maps.__iter__()

    def __len__(self):
        return self._maps.__len__()

    def __repr__(self):
        return f"<Carto Data: {self.__len__()} maps>"

    def add_signal_files(self, map_name=None):
        file_list = {file for file in self.data_source.listdir() if file.endswith('Point_Export.xml')}
        if map_name in self:
            file_list = {file for file in file_list if file.startswith(map_name)}
        elif map_name is not None:
            raise ValueError('Unknown map')

        for file in tqdm(file_list, ncols=50):
            self.add_signal_file(file)

    def add_signal_file(self, filename):
        if filename in self._integrated_files:
            return

        # We first need to read the xml file to read the timestamp and the name of the txt file with the signals
        with self.data_source.open(filename) as f:
            root = etree.parse(f).getroot()
            annotations = root.find('Annotations')
            start_ts = int(annotations.get('StartTime'))
            ecg = root.find('ECG')
            ecg_filename = ecg.get('FileName')

        # We then need to open the sig file to get the channel names and check if we already have the signals we need
        with self.data_source.open(ecg_filename) as f:
            f.readline()  # Version header
            gain_line = f.readline().decode('utf-8')
            gain = float(re.match(r'Raw ECG to MV \(gain\) = (?P<gain>[\d.]+)', gain_line)['gain'])
            mapping_chan_line = f.readline().decode('utf-8')
            # mapping_chan = re.match(r'Unipolar Mapping Channel=(?P<uni>[\w-]+) '
            #                         r'Bipolar Mapping Channel=(?P<bi>[\w-]+)'
            #                         r'( Reference Channel=(?P<ref>[\w-]+))?', mapping_chan_line)
            #
            # uni_chan, bi_chan, ref_chan = mapping_chan['uni'], mapping_chan['bi'], mapping_chan['ref']
            # uni_chan = uni_chan.replace('_', ':')
            # bi_chan = bi_chan.replace('_', ':')
            # ref_chan = ref_chan.replace('_', ':')

            labels_line = f.readline().decode('utf-8')
            labels = [re.match(r'(?P<name>[\w_-]+)\((?P<id>\d+)\)', label)['name'] for label in labels_line.split()]
            labels = [label.replace('_', ' ') for label in labels]

            # Now check we have the necessary channels info
            if all([label in self._signals and
                    self._signals[label].defined(start_ts, start_ts + CARTO_SIGNAL_FILE_LENGTH)
                    for label in labels]):
                self._integrated_files.add(filename)
                return

            # If we don't, we need to add the channels to the dataset
            # Get the signals
            signals = np.loadtxt(f).transpose()
            signals *= gain

        # Store them
        for label, signal in zip(labels, signals):
            if label not in self._signals:
                self._signals[label] = Channel(label)
            self._signals[label].add(signal, start_ts)
        self._integrated_files.add(filename)

    @property
    def data_source(self):
        return self._data_source

    @property
    def name(self):
        return self._name

    @property
    def sig(self):
        return self._signals

    @property
    def tags(self):
        return self._tags

    def signal_from_point(self, point: Point, interval: slice = slice(-500, 500)):
        if not isinstance(interval, slice):
            raise TypeError('Input interval must be a slice.')
        if interval.start is None or interval.stop is None:
            raise ValueError('Interval start and stop points must be defined.')

        offset = point.ref.ts + point.ref.ref_ts_delay
        offset_interval = slice(offset + interval.start, offset + interval.stop, interval.step)

        # Make sure we already integrated the signal file to our data
        self.add_signal_file(point.ref.filename)

        results = {
            't': np.arange(interval.start, interval.stop, interval.step),
            'uni': self.sig[point.ref.uni_chan][offset_interval],
            'bi': self.sig[point.ref.bi_chan][offset_interval],
        }

        if point.ref.ref_chan is not None:
            results['ref'] = self.sig[point.ref.ref_chan][offset_interval]

        return results

    def plot_point(self, point: Point, *args, **kwargs):
        try:
            from matplotlib import pyplot as plt
        except ImportError:
            raise ImportError('Matplotlib must be installed to plot signals. '
                              'Use pip install carto_reader[viz] to install visualization dependencies.')

        # Get signal
        sigs = self.signal_from_point(point, *args, **kwargs)

        uni = plt.plot(sigs['t'], sigs['uni'], 'r')
        bi = plt.plot(sigs['t'], sigs['bi'], 'b')
        try:
            ref = plt.plot(sigs['t'], sigs['ref'], 'g')
            plt.legend([f'Unipolar ({point.ref.uni_chan})',
                        f'Bipolar ({point.ref.bi_chan})',
                        f'Ref ({point.ref.ref_chan})'])
        except KeyError:
            plt.legend([f'Unipolar ({point.ref.uni_chan})', f'Bipolar ({point.ref.bi_chan})'])

        plt.axvline(point.ref.lat_ts_delay - point.ref.ref_ts_delay, c='k', ls='--')